package ctbrec.sites.cam4;

import static ctbrec.Model.State.*;
import static ctbrec.io.HttpConstants.*;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.ExecutionException;

import org.json.JSONArray;
import org.json.JSONObject;
import org.jsoup.nodes.Element;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.iheartradio.m3u8.Encoding;
import com.iheartradio.m3u8.Format;
import com.iheartradio.m3u8.ParseException;
import com.iheartradio.m3u8.ParsingMode;
import com.iheartradio.m3u8.PlaylistException;
import com.iheartradio.m3u8.PlaylistParser;
import com.iheartradio.m3u8.data.MasterPlaylist;
import com.iheartradio.m3u8.data.Playlist;
import com.iheartradio.m3u8.data.PlaylistData;
import com.iheartradio.m3u8.data.StreamInfo;

import ctbrec.AbstractModel;
import ctbrec.Config;
import ctbrec.io.HtmlParser;
import ctbrec.io.HttpException;
import ctbrec.recorder.download.StreamSource;
import okhttp3.FormBody;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;

public class Cam4Model extends AbstractModel {

    private static final Logger LOG = LoggerFactory.getLogger(Cam4Model.class);
    private String playlistUrl;
    private int[] resolution = null;
    private boolean privateRoom = false;

    @Override
    public boolean isOnline(boolean ignoreCache) throws IOException, ExecutionException, InterruptedException {
        if (ignoreCache || onlineState == UNKNOWN) {
            try {
                loadModelDetails();
            } catch (ModelDetailsEmptyException e) {
                return false;
            }
        }
        return onlineState == ONLINE && !privateRoom && playlistUrl != null && !playlistUrl.isEmpty();
    }

    private void loadModelDetails() throws IOException, ModelDetailsEmptyException {
        String url = site.getBaseUrl() + "/getBroadcasting?usernames=" + getName();
        LOG.trace("Loading model details {}", url);
        Request req = new Request.Builder().url(url).build();
        try (Response response = site.getHttpClient().execute(req)) {
            if (response.isSuccessful()) {
                JSONArray json = new JSONArray(response.body().string());
                if (json.length() == 0) {
                    onlineState = OFFLINE;
                    throw new ModelDetailsEmptyException("Model details are empty");
                }
                JSONObject details = json.getJSONObject(0);
                String showType = details.getString("showType");
                setOnlineStateByShowType(showType);
                playlistUrl = details.getString("hlsPreviewUrl");
                privateRoom = details.getBoolean("privateRoom");
                if (privateRoom) {
                    onlineState = PRIVATE;
                }
                if (details.has("resolution")) {
                    String res = details.getString("resolution");
                    String[] tokens = res.split(":");
                    resolution = new int[] { Integer.parseInt(tokens[0]), Integer.parseInt(tokens[1]) };
                }
            } else {
                throw new HttpException(response.code(), response.message());
            }
        }
    }

    public void setOnlineStateByShowType(String showType) {
        switch(showType) {
        case "NORMAL":
        case "GROUP_SHOW_SELLING_TICKETS":
            onlineState = ONLINE;
            break;
        case "PRIVATE_SHOW":
            onlineState = PRIVATE;
            break;
        case "GROUP_SHOW":
            onlineState = GROUP;
            break;
        case "OFFLINE":
            onlineState = OFFLINE;
            break;
        default:
            LOG.debug("Unknown show type {}", showType);
            onlineState = UNKNOWN;
        }

    }

    @Override
    public State getOnlineState(boolean failFast) throws IOException, ExecutionException {
        if(failFast) {
            return onlineState;
        } else {
            if(onlineState == UNKNOWN) {
                try {
                    loadModelDetails();
                } catch (ModelDetailsEmptyException e) {
                    LOG.warn("Couldn't load model details {}", e.getMessage());
                }
            }
            return onlineState;
        }
    }

    private String getPlaylistUrl() throws IOException {
        if(playlistUrl == null || playlistUrl.trim().isEmpty()) {
            try {
                loadModelDetails();
                if (playlistUrl == null) {
                    throw new IOException("Couldn't determine playlist url");
                }
            } catch (ModelDetailsEmptyException e) {
                throw new IOException(e);
            }
        }
        return playlistUrl;
    }

    @Override
    public List<StreamSource> getStreamSources() throws IOException, ExecutionException, ParseException, PlaylistException {
        MasterPlaylist masterPlaylist = getMasterPlaylist();
        List<StreamSource> sources = new ArrayList<>();
        for (PlaylistData playlist : masterPlaylist.getPlaylists()) {
            if (playlist.hasStreamInfo()) {
                StreamSource src = new StreamSource();
                src.bandwidth = playlist.getStreamInfo().getBandwidth();
                src.height = Optional.ofNullable(playlist.getStreamInfo()).map(StreamInfo::getResolution).map(res -> res.height).orElse(0);
                String masterUrl = getPlaylistUrl();
                String baseUrl = masterUrl.substring(0, masterUrl.lastIndexOf('/') + 1);
                String segmentUri = baseUrl + playlist.getUri();
                src.mediaPlaylistUrl = segmentUri;
                LOG.trace("Media playlist {}", src.mediaPlaylistUrl);
                sources.add(src);
            }
        }
        return sources;
    }

    private MasterPlaylist getMasterPlaylist() throws IOException, ParseException, PlaylistException {
        LOG.debug("Loading master playlist [{}]", getPlaylistUrl());
        Request req = new Request.Builder().url(getPlaylistUrl()).build();

        try (Response response = site.getHttpClient().execute(req)) {
            if (response.isSuccessful()) {
                InputStream inputStream = response.body().byteStream();
                PlaylistParser parser = new PlaylistParser(inputStream, Format.EXT_M3U, Encoding.UTF_8, ParsingMode.LENIENT);
                Playlist playlist = parser.parse();
                MasterPlaylist master = playlist.getMasterPlaylist();
                return master;
            } else {
                throw new HttpException(response.code(), "Couldn't download HLS playlist");
            }
        }
    }

    @Override
    public void invalidateCacheEntries() {
        resolution = null;
        playlistUrl = null;
    }

    @Override
    public void receiveTip(Double tokens) throws IOException {
        throw new RuntimeException("Not implemented for Cam4, yet");
    }

    @Override
    public int[] getStreamResolution(boolean failFast) throws ExecutionException {
        if(resolution == null) {
            if(failFast) {
                return new int[2];
            } else {
                try {
                    if(onlineState != OFFLINE) {
                        loadModelDetails();
                    } else {
                        resolution = new int[2];
                    }
                } catch (Exception e) {
                    throw new ExecutionException(e);
                }
            }
        }
        return resolution;
    }

    @Override
    public boolean follow() throws IOException {
        String url = site.getBaseUrl() + "/profiles/addFriendFavorite?action=addFavorite&object=" + getName() + "&_=" + System.currentTimeMillis();
        Request req = new Request.Builder()
                .url(url)
                .addHeader(X_REQUESTED_WITH, XML_HTTP_REQUEST)
                .build();
        try (Response response = site.getHttpClient().execute(req)) {
            return response.isSuccessful();
        }
    }

    @Override
    public boolean unfollow() throws IOException {
        // get model user id
        String url = site.getBaseUrl() + '/' + getName();
        Request req = new Request.Builder()
                .url(url)
                .addHeader(X_REQUESTED_WITH, XML_HTTP_REQUEST)
                .build();

        // we have to use a client without any cam4 cookies here, otherwise
        // this request is redirected to the login page. no idea why
        try (Response response = site.getRecorder().getHttpClient().execute(req)) {
            if (response.isSuccessful()) {
                String content = response.body().string();
                String broadCasterId = null;
                try {
                    Element tag = HtmlParser.getTag(content, "input[name=\"broadcasterId\"]");
                    broadCasterId = tag.attr("value");
                } catch (Exception e) {
                    LOG.debug(content);
                    throw new IOException(e);
                }

                // send unfollow request
                String username = Config.getInstance().getSettings().cam4Username;
                url = site.getBaseUrl() + '/' + username + "/edit/friends_favorites";
                RequestBody body = new FormBody.Builder()
                        .add("deleteFavorites", broadCasterId)
                        .add("simpleresult", "true")
                        .build();
                req = new Request.Builder()
                        .url(url)
                        .post(body)
                        .addHeader(X_REQUESTED_WITH, XML_HTTP_REQUEST)
                        .build();
                try (Response resp = site.getHttpClient().execute(req)) {
                    if (resp.isSuccessful()) {
                        return Objects.equals(resp.body().string(), "Ok");
                    } else {
                        return false;
                    }
                }
            } else {
                return false;
            }
        }
    }

    public void setPlaylistUrl(String playlistUrl) {
        this.playlistUrl = playlistUrl;
    }

    public class ModelDetailsEmptyException extends Exception {
        public ModelDetailsEmptyException(String msg) {
            super(msg);
        }
    }
}
